

#include <windows.h>
//#include <stdio.h>
#include "classparser.h"


extern HANDLE heap;


struct classfile parseClassFromFile(HANDLE hFile){

	DWORD d;

	struct classfile mainstr;

	ReadFileBigEndian(hFile, &(mainstr.magic), 4, &d, NULL);
	ReadFileBigEndian(hFile, (void*)&mainstr.minor_version, 2, &d, NULL);
	ReadFileBigEndian(hFile, (void*)&mainstr.major_version, 2, &d, NULL);
	ReadFileBigEndian(hFile, (void*)&mainstr.constant_pool_count, 2, &d, NULL);

	mainstr.constant_pool = (cp_info*)HeapAlloc(heap,0,mainstr.constant_pool_count*sizeof(cp_info));

	bool bDoubleItem = false;
	unsigned char u1;
	cp_info cptemp;
	for(int i=1; i<mainstr.constant_pool_count;i++){
		if(bDoubleItem){
			mainstr.constant_pool[i].tag = 0;
			bDoubleItem=false;
			continue;	
		}

		ReadFile(hFile, (void*)&u1, 1, &d, NULL);
		cptemp.tag = u1;

		switch(u1){
		case CONSTANT_Utf8:
			ReadFileBigEndian(hFile, &cptemp.value.u2, 2, &d, NULL);
			if(cptemp.value.u2>0){
				cptemp.value1.string = (unsigned char *)HeapAlloc(heap,0,cptemp.value.u2);
				ReadFile(hFile, cptemp.value1.string, cptemp.value.u2, &d, NULL);
			} else {
				cptemp.value1.string = 0;
			}
			break;
		case CONSTANT_Integer:
			ReadFileBigEndian(hFile, &cptemp.value.u4, 4, &d, NULL);
			break;
		case CONSTANT_Float:
			ReadFileBigEndian(hFile, &cptemp.value.f, 4, &d, NULL);
			break;
		case CONSTANT_Long:
			ReadFileBigEndian(hFile, &cptemp.value.u8, 8, &d, NULL);
			bDoubleItem=true; 
			break;
		case CONSTANT_Double:
			ReadFileBigEndian(hFile, &cptemp.value.d, 8, &d, NULL);
			bDoubleItem=true;
			break;
		case CONSTANT_Class:
		case CONSTANT_String:
			ReadFileBigEndian(hFile, &cptemp.value.u2, 2, &d, NULL);
			break;
		case CONSTANT_Fieldref:
		case CONSTANT_Methodref:
		case CONSTANT_InterfaceMethodref:
		case CONSTANT_NameAndType:
			ReadFileBigEndian(hFile, &cptemp.value.u2, 2, &d, NULL);
			ReadFileBigEndian(hFile, &cptemp.value1.u2, 2, &d, NULL);
			break;
		}

		mainstr.constant_pool[i]=cptemp;
		cptemp.value.u8 = 0;
		cptemp.value1.u2 = 0;

	}


	ReadFileBigEndian(hFile, &mainstr.access_flags, 2, &d, NULL);
	ReadFileBigEndian(hFile, &mainstr.this_class, 2, &d, NULL);
	ReadFileBigEndian(hFile, &mainstr.super_class, 2, &d, NULL);
	ReadFileBigEndian(hFile, &mainstr.interfaces_count, 2, &d, NULL);

	if(mainstr.interfaces_count>0){
		mainstr.interfaces = (unsigned short *)HeapAlloc(heap,0,mainstr.interfaces_count*2);
		for(int i=0;i<mainstr.interfaces_count;i++){
			ReadFileBigEndian(hFile, &(mainstr.interfaces[i]), 2, &d, NULL);
		}	
	}

	ReadFileBigEndian(hFile, &mainstr.fields_count, 2, &d, NULL);

	if(mainstr.fields_count>0){
		mainstr.fields = (field_info *)HeapAlloc(heap,0,mainstr.fields_count*sizeof(field_info));
		
		field_info ftemp;
		for(int i=0;i<mainstr.fields_count;i++){
			ReadFileBigEndian(hFile, &ftemp.access_flags, 2, &d, NULL);
			ReadFileBigEndian(hFile, &ftemp.name_index, 2, &d, NULL);
			ReadFileBigEndian(hFile, &ftemp.descriptor_index, 2, &d, NULL);
			ReadFileBigEndian(hFile, &ftemp.attributes_count, 2, &d, NULL);

			if(ftemp.attributes_count>0){
				parseAttributes(hFile,&ftemp.attributes,ftemp.attributes_count);
			} else {
				ftemp.attributes = 0;
			}
			mainstr.fields[i]=ftemp;
		}
	}


	ReadFileBigEndian(hFile, &mainstr.methods_count, 2, &d, NULL);

	if(mainstr.methods_count>0){
		mainstr.methods = (method_info *)HeapAlloc(heap,0,mainstr.methods_count*sizeof(method_info));
		
		method_info mtemp;
		for(int i=0;i<mainstr.methods_count;i++){
			ReadFileBigEndian(hFile, &mtemp.access_flags, 2, &d, NULL);
			ReadFileBigEndian(hFile, &mtemp.name_index, 2, &d, NULL);
			ReadFileBigEndian(hFile, &mtemp.descriptor_index, 2, &d, NULL);
			ReadFileBigEndian(hFile, &mtemp.attributes_count, 2, &d, NULL);

			if(mtemp.attributes_count>0){
				parseAttributes(hFile,&mtemp.attributes,mtemp.attributes_count);
			} else {
				mtemp.attributes = 0;
			}
			mainstr.methods[i]=mtemp;
		}
	}

	ReadFileBigEndian(hFile, &mainstr.attributes_count, 2, &d, NULL);

	if(mainstr.attributes_count>0){
		parseAttributes(hFile,&mainstr.attributes,mainstr.attributes_count);
	} 

	return mainstr;
}


void parseAttributes(HANDLE hFile, attribute_info** attr, int attr_count)
{
	DWORD d;
	*attr = (attribute_info *)HeapAlloc(heap,0,attr_count*sizeof(attribute_info));

	attribute_info atemp;
	for(int j=0;j<attr_count;j++){
		ReadFileBigEndian(hFile, &atemp.attribute_name_index, 2, &d, NULL);
		ReadFileBigEndian(hFile, &atemp.attribute_length, 4, &d, NULL);
		if(atemp.attribute_length>0){
			atemp.info = (unsigned char *)HeapAlloc(heap,0,atemp.attribute_length);
			ReadFile(hFile, atemp.info, atemp.attribute_length, &d, NULL);
		} else {
			atemp.info = 0;
		}
		(*attr)[j]=atemp;
	}
}



void writeClassToFile(HANDLE hFile, struct classfile mainstr)
{

	DWORD d;

	WriteFileBigEndian(hFile, &mainstr.magic, 4, &d, NULL);
	WriteFileBigEndian(hFile, &mainstr.minor_version, 2, &d, NULL);
	WriteFileBigEndian(hFile, &mainstr.major_version, 2, &d, NULL);
	WriteFileBigEndian(hFile, &mainstr.constant_pool_count, 2, &d, NULL);

	for(int i=1; i<mainstr.constant_pool_count;i++){
		cp_info cptemp = mainstr.constant_pool[i];
		unsigned char u1 = cptemp.tag;
		
		if(u1){
			WriteFile(hFile, &u1 , 1, &d, NULL);
 
			switch(u1){
			case CONSTANT_Utf8:
				WriteFileBigEndian(hFile, &cptemp.value.u2, 2, &d, NULL);
				if(cptemp.value.u2>0)WriteFile(hFile, cptemp.value1.string, cptemp.value.u2, &d, NULL);
				break;
			case CONSTANT_Integer:
				WriteFileBigEndian(hFile, &cptemp.value.u4, 4, &d, NULL);
				break;
			case CONSTANT_Float:
				WriteFileBigEndian(hFile, &cptemp.value.f, 4, &d, NULL);
				break;
			case CONSTANT_Long:
				WriteFileBigEndian(hFile, &cptemp.value.u8, 8, &d, NULL);
				break;
			case CONSTANT_Double:
				WriteFileBigEndian(hFile, &cptemp.value.d, 8, &d, NULL);
				break;
			case CONSTANT_Class:
			case CONSTANT_String:
				WriteFileBigEndian(hFile, &cptemp.value.u2, 2, &d, NULL);
				break;
			case CONSTANT_Fieldref:
			case CONSTANT_Methodref:
			case CONSTANT_InterfaceMethodref:
			case CONSTANT_NameAndType:
				WriteFileBigEndian(hFile, &cptemp.value.u2, 2, &d, NULL);
				WriteFileBigEndian(hFile, &cptemp.value1.u2, 2, &d, NULL);
				break;
			}
		}
	}

	WriteFileBigEndian(hFile, &mainstr.access_flags, 2, &d, NULL);
	WriteFileBigEndian(hFile, &mainstr.this_class, 2, &d, NULL);
	WriteFileBigEndian(hFile, &mainstr.super_class, 2, &d, NULL);
	WriteFileBigEndian(hFile, &mainstr.interfaces_count, 2, &d, NULL);


	if(mainstr.interfaces_count>0){
		for(int i=0;i<mainstr.interfaces_count;i++){
			WriteFileBigEndian(hFile, &(mainstr.interfaces[i]), 2, &d, NULL);
		}	
	}

	WriteFileBigEndian(hFile, &mainstr.fields_count, 2, &d, NULL);

	if(mainstr.fields_count>0){
		
		for(int i=0;i<mainstr.fields_count;i++){
			field_info ftemp = mainstr.fields[i];
			WriteFileBigEndian(hFile, &ftemp.access_flags, 2, &d, NULL);
			WriteFileBigEndian(hFile, &ftemp.name_index, 2, &d, NULL);
			WriteFileBigEndian(hFile, &ftemp.descriptor_index, 2, &d, NULL);
			WriteFileBigEndian(hFile, &ftemp.attributes_count, 2, &d, NULL);

			if(ftemp.attributes_count>0){
				writeAttributes(hFile,ftemp.attributes,ftemp.attributes_count);
			}
		}
	}


	WriteFileBigEndian(hFile, &mainstr.methods_count, 2, &d, NULL);

	if(mainstr.methods_count>0){
		for(int i=0;i<mainstr.methods_count;i++){
			method_info mtemp = mainstr.methods[i];
			WriteFileBigEndian(hFile, &(mtemp.access_flags), 2, &d, NULL);
			WriteFileBigEndian(hFile, &(mtemp.name_index), 2, &d, NULL);
			WriteFileBigEndian(hFile, &(mtemp.descriptor_index), 2, &d, NULL);
			WriteFileBigEndian(hFile, &(mtemp.attributes_count), 2, &d, NULL);

			if(mtemp.attributes_count>0){
				writeAttributes(hFile,mtemp.attributes,mtemp.attributes_count);
			}
		}
	}


	WriteFileBigEndian(hFile, &mainstr.attributes_count, 2, &d, NULL);

	if(mainstr.attributes_count>0){
		writeAttributes(hFile,mainstr.attributes,mainstr.attributes_count);
	}


}

void writeAttributes(HANDLE hFile, attribute_info* attr, int attr_count)
{
	DWORD d;
	for(int j=0;j<attr_count;j++){
		attribute_info atemp = attr[j];
		WriteFileBigEndian(hFile, &atemp.attribute_name_index, 2, &d, NULL);
		WriteFileBigEndian(hFile, &atemp.attribute_length, 4, &d, NULL);
		if(atemp.attribute_length>0)WriteFile(hFile, atemp.info, atemp.attribute_length, &d, NULL);
	}
}


void freeClassStruct(struct classfile mainstr) 
{

	if(mainstr.constant_pool_count>0){
		for(int i=1;i<mainstr.constant_pool_count;i++){
			if(mainstr.constant_pool[i].tag == 1){
				HeapFree(heap,0,mainstr.constant_pool[i].value1.string);
			}
		}

		HeapFree(heap,0,mainstr.constant_pool);
	}

	if(mainstr.interfaces_count>0){
		HeapFree(heap,0,mainstr.interfaces);
	}

	if(mainstr.attributes_count>0){
		for(int i=0;i<mainstr.attributes_count;i++){
			if(mainstr.attributes[i].attribute_length>0){
				HeapFree(heap,0,mainstr.attributes[i].info);
			}
		}

		HeapFree(heap,0,mainstr.attributes);
	}

	if(mainstr.fields_count>0){
		for(int i=0;i<mainstr.fields_count;i++){
			if(mainstr.fields[i].attributes_count>0){
				for(int j=0;j<mainstr.fields[i].attributes_count;j++){
					if(mainstr.fields[i].attributes[j].attribute_length>0){
						HeapFree(heap,0,mainstr.fields[i].attributes[j].info);
					}
				}

				HeapFree(heap,0,mainstr.fields[i].attributes);
			}
		}

		HeapFree(heap,0,mainstr.fields);
	}

	if(mainstr.methods_count>0){
		for(int i=0;i<mainstr.methods_count;i++){
			if(mainstr.methods[i].attributes_count>0){
				for(int j=0;j<mainstr.methods[i].attributes_count;j++){
					if(mainstr.methods[i].attributes[j].attribute_length>0){
						HeapFree(heap,0,mainstr.methods[i].attributes[j].info);
					}
				}

				HeapFree(heap,0,mainstr.methods[i].attributes);
			}
		}

		HeapFree(heap,0,mainstr.methods);
	}


}


BOOL ReadFileBigEndian(HANDLE hFile, LPVOID lpBuffer, DWORD nNumberOfBytesToRead, LPDWORD lpNumberOfBytesRead, LPOVERLAPPED lpOverlapped)
{
	BOOL res = ReadFile(hFile, lpBuffer, nNumberOfBytesToRead, lpNumberOfBytesRead, lpOverlapped);
	if(res){
		invertBytesEx(lpBuffer,nNumberOfBytesToRead);
	}
	return res;	
}

BOOL WriteFileBigEndian(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped)
{
	unsigned char* c = (unsigned char*)HeapAlloc(heap,0,nNumberOfBytesToWrite);
	CopyMemory(c,lpBuffer,nNumberOfBytesToWrite);

	invertBytesEx(c,nNumberOfBytesToWrite);

	BOOL res = WriteFile(hFile, c, nNumberOfBytesToWrite, lpNumberOfBytesWritten, lpOverlapped);
	
	HeapFree(heap,0,c);
	return res;
}

void invertBytesEx(void* v, DWORD len)
{
	int l = len/2;
	char c;
	char* ve = (char*)v+len-1; 

	while(l--){
		c=*(char*)v;
		*(char*)v=*ve;
		*ve=c;

		v=(char *)v+1;
		ve--;
	}

}

